const terminalBody = document.getElementById('terminal-body');
const commandInput = document.getElementById('command-input');
const autocompleteList = document.getElementById('autocomplete-list');
const sendButton = document.getElementById('send-btn');
const formatButton = document.getElementById('format-btn');
let formatButtonHidden = false;
let socket;
let accumulatedResults = [];
let receivingBatches = false;

// Commands
const commands = ['QUERY', 'SET', 'RENAME', 'DEL', 'INCR', 'DECR','COUNT', 'BACKUP', 'INDICES LIST', 'INDICES GET', 'INDICES CREATE', 'INDICES DEL', 'INDICES FLUSH', 'SCHEDULE', 'UNSUB', 'SUB', 'SUBLIST', 'CONFIG', 'CHECKUPDATE', 'FLUSHCACHE', 'FLUSHALL'];
const subCommands = ['WHERE', 'BETWEEN', 'ORDERBY(key,ASC/DESC)', 'GROUPBY(key)', 'LIMIT()', 'INCLUDE(key1,key2)', 'EXCLUDE(key1,key2)'];

// Connect to WebSocket server
function connectToWebSocket() {
    const uri = 'ws://127.0.0.1:6446';
    if (typeof socket !== 'undefined' && socket.readyState !== WebSocket.CLOSED) {
        console.log('WebSocket connection already open.');
        return;
    }
    socket = new WebSocket(uri);

    socket.onopen = function () {
        const authData = JSON.stringify({
            username: '',
            password: ''
        });
        socket.send(authData);
        displayResponse('Attempting to authenticate with MginDB server...');
        sendButton.style.display = '';
        commandInput.disabled = false;
        commandInput.focus();
    };

    socket.onmessage = function (event) {
        handleWebSocketMessage(event.data);
    };

    socket.onclose = function () {
        disableInputAndNotifyUser("MginDB server connection closed. Reconnecting...");
        sendButton.style.display = 'none';
        // Attempt to reconnect every 10 seconds
        setTimeout(connectToWebSocket, 5000);
    };

    socket.onerror = function (error) {
        console.error('WebSocket error:', error);
        sendButton.style.display = 'none';
        disableInputAndNotifyUser("Error connecting to MginDB server. Reconnecting...");
        // Attempt to reconnect every 10 seconds
        setTimeout(connectToWebSocket, 5000);
    };
}

// Helper function to disable input and display a message in the terminal
function disableInputAndNotifyUser(message) {
    const commandInput = document.getElementById('command-input');
    commandInput.disabled = true; // Disable the input field to prevent user interaction
    displayResponse(message); // Display the error message in the terminal
}

function handleWebSocketMessage(data) {
    if (receivingBatches) {
        const parsedData = JSON.parse(data);
        accumulatedResults.push(...parsedData);
    } else {
        if (data.includes("Results sent in batches via WebSocket.")) {
            receivingBatches = true;
            accumulatedResults = [];
        } else {
            displayResponse(data);
        }
    }
}

function enforceKeywordUppercase() {
    let userInput = commandInput.value;
    const inputParts = userInput.split(' ');
    const allKeywords = [...commands, ...subCommands.map(item => item.split('(')[0])]; // Extract base command from subcommands

    // Transform only the parts of the input that are recognized commands or subcommands
    const transformedInput = inputParts.map(part => {
        const partUpper = part.toUpperCase();
        if (allKeywords.includes(partUpper)) {
            return partUpper;  // Convert recognized keywords to uppercase
        }
        return part;  // Leave other text as is
    }).join(' ');

    // Update the input field if changes were made
    if (userInput !== transformedInput) {
        commandInput.value = transformedInput;
        commandInput.selectionStart = commandInput.selectionEnd = transformedInput.length;
    }
}

function updateCommandHint() {
    const userInput = commandInput.value.trim();
    const inputParts = userInput.split(' ');
    const commandPrefix = inputParts[0].toUpperCase();

    // Commands for which no hint is displayed
    const noHintCommands = ['BACKUP', 'SUBLIST', 'CHECKUPDATE', 'FLUSHCACHE', 'FLUSHALL'];

    // Reference to the hint element
    const hintElement = document.getElementById('input-hint');
    hintElement.style.left = '';
    let hintContent = '';

    // Function to manage hints by command
    const hintsByCommand = {
        'QUERY': () => {
            const whereIndex = inputParts.indexOf('WHERE');
            if (whereIndex >= 0 && inputParts.length === whereIndex + 1) {
                hintElement.style.left = '190px';
                return 'criteria = != / <> / LIKE / BETWEEN value';
            } else if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey';
            }
            return '';
        },
        'SET': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey value';
            }
            return '';
        },
        'RENAME': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '80px';
                return 'key / key:subkey TO new_name';
            }
            return '';
        },
        'DEL': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey';
            }
            return '';
        },
        'INCR': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey value';
            }
            return '';
        },
        'DECR': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey value';
            }
            return '';
        },
        'COUNT': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key / key:subkey';
            }
            return '';
        },
        'SCHEDULE': () => {
            if (inputParts.includes('SHOW') && (inputParts.length === 2 || inputParts.length === 1)) {
                return 'ALL or Cron format e.g., * * * * * or key / key:subkey';
            } else if (inputParts.includes('ADD') && (inputParts.length === 2 || inputParts.length === 1)) {
                return '* * * * * COMMAND(SET key value)';
            } else if (inputParts.includes('DEL') && (inputParts.length === 2 || inputParts.length === 1)) {
                return 'key / key:subkey';
            } else if (inputParts.includes('FLUSH') && (inputParts.length === 2 || inputParts.length === 1)) {
                return 'cron format e.g., * * * * * or key / key:subkey';
            } else if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                return 'Specify SHOW, ADD, SET, DEL, FLUSH';
            }
            return '';
        },
        'SUB': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key:* / key:subkey:* or key:value / key:subkey:value';
            }
            return '';
        },
        'UNSUB': () => {
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                hintElement.style.left = '70px';
                return 'key:* / key:subkey:* or key:value / key:subkey:value';
            }
            return '';
        },
        'CONFIG': () => {
            if (inputParts.includes('SET') && inputParts.includes('REPLICATION_AUTHORIZED_SLAVES') && (inputParts.length === 3 || inputParts.length === 2)) {
                hintElement.style.left = '200px';
                return 'ADD/DEL slaveip or domain';
            } else if (inputParts.includes('SET') && inputParts.includes('SHARDS') && (inputParts.length === 3 || inputParts.length === 2)) {
                hintElement.style.left = '200px';
                return 'ADD/DEL server or domain';
            } else if (inputParts.includes('SET') && (inputParts.includes('SHARDING') || inputParts.includes('SCHEDULER')) && (inputParts.length === 3 || inputParts.length === 2)) {
                hintElement.style.left = '200px';
                return '0 to activate / 1 to deactivate';
            } else if (inputParts.includes('SET') && (inputParts.includes('REPLICATION_TYPE') || inputParts.includes('SHARDING_TYPE')) && (inputParts.length === 3 || inputParts.length === 2)) {
                hintElement.style.left = '220px';
                return 'MASTER/SLAVE';
            }
            
            if (inputParts.includes('SET') && (inputParts.length === 2 || inputParts.length === 1)) {
                return 'key / key:subkey value';
            } else if (inputParts.includes('DEL') && (inputParts.length === 2 || inputParts.length === 1)) {
                return 'key / key:subkey';
            }
            
            if (inputParts.length === 1 || (inputParts.length === 2 && inputParts[1] === '')) {
                return 'Specify SHOW, SET, DEL, SET REPLICATION_AUTHORIZED_SLAVES, SET SHARDS';
            }

            return '';
        },
        // Other commands...
    };

    // Check if the first word is a valid command and not a no-hint command
    if (commands.includes(commandPrefix) && !noHintCommands.includes(commandPrefix)) {
        const hintFunction = hintsByCommand[commandPrefix];
        hintContent = hintFunction ? hintFunction() : 'key or key:subkey';  // Default hint
    }

    // Update the hint in the DOM
    hintElement.textContent = hintContent;
}

function autocompleteCommand() {
    const userInput = commandInput.value.toUpperCase();
    // Splitting input to detect primary command and subsequent text
    const inputParts = userInput.split(' ');
    const primaryCommand = inputParts[0];
    const restInput = inputParts.slice(1).join(' ');

    let suggestions = [];

    if (primaryCommand === 'QUERY') {
        // User is using QUERY command, suggest only subcommands
        if (restInput.length > 0) {
            suggestions = subCommands.filter(subCommand => 
                subCommand.startsWith(restInput.split(' ').pop()) && !userInput.includes(subCommand.split('(')[0].toUpperCase())
            );
        } else {
            // No additional input, suggest starting subcommands
            suggestions = subCommands.filter(subCommand => 
                !userInput.includes(subCommand.split('(')[0].toUpperCase())
            );
        }
    } else {
        // User entered another command, suggest top-level commands
        suggestions = commands.filter(command => 
            command.startsWith(primaryCommand) && !userInput.includes(command.toUpperCase())
        );
    }

    updateAutocompleteList(suggestions);
}

function updateAutocompleteList(suggestions) {
    autocompleteList.innerHTML = '';
    if (suggestions.length > 0) {
        suggestions.forEach(suggestion => {
            const listItem = document.createElement('div');
            listItem.textContent = suggestion;
            listItem.classList.add('autocomplete-list-item');
            listItem.addEventListener('click', () => {
                const currentInput = commandInput.value.split(' ');
                if (commands.includes(currentInput[0])) {
                    // Replace or add subcommand in input
                    currentInput[currentInput.length - 1] = suggestion;
                    commandInput.value = currentInput.join(' ') + ' ';
                } else {
                    if(suggestion !== 'BACKUP'
                        && suggestion !== 'CONFIG SHOW'
                        && suggestion !== 'CONFIG SET'
                        && suggestion !== 'CONFIG DEL'
                        && suggestion !== 'CHECKUPDATE'
                        && suggestion !== 'SUBLIST') {
                        document.getElementById('input-hint').innerHTML = 'key or key:subkey';
                    }
                    commandInput.value = suggestion + ' ';
                }
                commandInput.focus();
                autocompleteList.innerHTML = '';
            });
            autocompleteList.appendChild(listItem);
        });
        autocompleteList.style.display = 'block';
    } else {
        autocompleteList.style.display = 'none';
    }
}

// Function to safely escape HTML entities to prevent XSS
function escapeHTML(text) {
    var element = document.createElement('div');
    if (text) {
        element.innerText = text;
        element.textContent = text;
    }
    return element.innerHTML;
}

// Function to put command back into the input field
function putCommandInInput(command) {
    const commandInput = document.getElementById('command-input');
    commandInput.value = command; // Set the command as the input's value
    commandInput.focus(); // Optionally, focus the input field
}

// Send command function
function sendCommand() {
    const command = commandInput.value.trim();
    formatButtonHidden = false;

    if(command === 'clear') {
        clearTerminal();
        commandInput.value = '';
        return false;
    }

    if(!command.startsWith('QUERY')) {
        formatButtonHidden = true;
    }

    // Check if command starts with one of the prefix keywords
    const startsWithPrefix = commands.some(prefix => command.toUpperCase().startsWith(prefix));
    
    if (command === '' || !startsWithPrefix) return;

    // Display the command in the terminal
    displayCommand(command);

    // Send command to WebSocket server
    socket.send(command);

    // Clear command input after displaying the command
    commandInput.value = '';
    autocompleteList.innerHTML = '';
}

function displayCommand(command) {
    const commandElement = document.createElement('div');
    const escapedCommand = command.replace(/'/g, "\\'");
    commandElement.innerHTML = `<a href="#" class="command-link" onclick="putCommandInInput('${escapeHTML(escapedCommand)}'); return false;">> ${command}</a>`;
    commandElement.classList.add('terminal-message', 'command-message');
    terminalBody.appendChild(commandElement);
    terminalBody.scrollTop = terminalBody.scrollHeight;
}

function displayResponse(response) {
    const responseElement = document.createElement('div');
    responseElement.textContent = response;

    // Add a class to style the message
    responseElement.classList.add('terminal-message');

    // Append the message to the terminal body
    terminalBody.appendChild(responseElement);

    // Scroll to the bottom of the terminal body to keep the latest message visible
    terminalBody.scrollTop = terminalBody.scrollHeight;

    // Check if the response is JSON and display the format button if it is
    try {
        const jsonData = JSON.parse(response); // Attempt to parse the response as JSON
        if(!formatButtonHidden) {
            formatButton.style.display = '';
        } else {
            formatButton.style.display = 'none';
        }
    } catch (error) {
        formatButton.style.display = 'none';
    }
}

function clearTerminal() {
    formatButton.style.display = 'none';
    // Remove all child elements from terminal body
    while (terminalBody.firstChild) {
        terminalBody.removeChild(terminalBody.firstChild);
    }
}

function formatData() {
    formatButton.style.display = 'none';
    const messages = terminalBody.querySelectorAll('.terminal-message');
    const lastMessage = messages[messages.length - 1];
    const data = lastMessage.textContent.trim();

    try {
        const jsonData = JSON.parse(data);
        const formattedData = formatAsTable(jsonData);
        lastMessage.innerHTML = `<div class="table-wrapper">${formattedData}</div>`; // Wrap table inside a div
    } catch (error) {
        console.error('Error formatting data:', error);
    }
}

function formatAsTable(data) {
    // This variable checks if any grouped data has been processed
    let groupedDataProcessed = false;

    // Check if the input is an array
    if (Array.isArray(data)) {
        // Determine if it's grouped by checking the first item
        const isGrouped = data.length > 0 && !Array.isArray(data[0]) && typeof data[0] === 'object' && Object.keys(data[0]).every(key => Array.isArray(data[0][key]));

        if (isGrouped) {
            // Handle grouped data (array of grouped objects)
            let result = '';
            data.forEach(group => {
                Object.keys(group).forEach(groupKey => {
                    const groupData = group[groupKey];
                    if (Array.isArray(groupData)) {
                        // Group data is an array, format as a table
                        result += `<h3>${groupKey}</h3>${formatAsTable(groupData)}`;
                        groupedDataProcessed = true; // Mark that grouped data has been processed
                    }
                });
            });
            return result; // Only return if grouped data was indeed processed
        }
    }

    // If no grouped data was processed, handle as flat array or other types
    if (!groupedDataProcessed) {
        if (Array.isArray(data)) {
            const keys = Object.keys(data[0]);

            let table = '<table class="table"><thead><tr>';
            keys.forEach(key => {
                table += `<th>${key}</th>`;
            });
            table += '</tr></thead><tbody>';

            data.forEach(item => {
                table += '<tr>';
                keys.forEach(key => {
                    const value = item[key];
                    if (Array.isArray(value)) {
                        // If value is an array, create a collapsible tree
                        table += `<td><a href="#" class="expand-array" data-length="${value.length}">${value.length} ${value.length > 1 ? 'items' : 'item'}</a>`;
                        table += `<ol start="0" style="display: none;">`;
                        value.forEach(val => {
                            table += `<li>${val}</li>`;
                        });
                        table += `</ol></td>`;
                    } else if (typeof value === 'object' && value !== null) {
                        table += `<td>${formatAsTable(value)}</td>`;
                    } else {
                        table += `<td>${value !== undefined ? value : ''}</td>`;
                    }
                });
                table += '</tr>';
            });

            table += '</tbody></table>';
            return table;
        } else if (typeof data === 'object' && data !== null) {
            let result = '<table class="table"><thead><tr><th>Field</th><th>Value</th></tr></thead><tbody>';
            Object.entries(data).forEach(([key, value]) => {
                result += `<tr><td>${key}</td><td>${value}</td></tr>`;
            });
            result += '</tbody></table>';
            return result;
        } else {
            return 'Invalid data format'; // Handle non-object, non-array data formats
        }
    }
}

function handleInputEvents() {
    autocompleteCommand();
    updateCommandHint();
    enforceKeywordUppercase();
}

// Connect to WebSocket server when the page loads
window.onload = function () {
    let data = [];
    connectToWebSocket();

    // Focus on the input field when the page loads
    commandInput.focus();
    commandInput.addEventListener('input', handleInputEvents);

    // Press Enter key to send command
    commandInput.addEventListener('keydown', function (event) {
        if (event.key === 'Enter') {
            sendCommand();
        }
    });

    // Add event listeners to expand array links
    document.addEventListener('click', function(event) {
        if (event.target && event.target.classList.contains('expand-array')) {
            event.preventDefault();
            const ulElement = event.target.nextElementSibling;
            if (ulElement) {
                // Create overlay modal
                const modalOverlay = document.createElement('div');
                modalOverlay.classList.add('modal-overlay');
                modalOverlay.innerHTML = '<div class="modal-content"></div>';
                document.body.appendChild(modalOverlay);

                // Populate modal content with the existing ordered list from the table
                const modalContent = modalOverlay.querySelector('.modal-content');
                const clonedOlElement = ulElement.cloneNode(true); // Clone the ol element
                clonedOlElement.style.display = 'block';
                modalContent.appendChild(clonedOlElement);

                // Display modal overlay
                modalOverlay.style.display = 'block';

                // Close modal when clicking outside the content
                modalOverlay.addEventListener('click', function(event) {
                    if (!event.target.closest('.modal-content')) {
                        modalOverlay.style.display = 'none';
                    }
                });
            }
        }
    });

    data.forEach(item => {
        table += '<tr>';
        keys.forEach(key => {
            const value = item[key];
            if (Array.isArray(value)) {
                // If value is an array, create a collapsible tree
                table += `<td><a href="#" class="expand-array" data-length="${value.length}">${value.length} ${value.length > 1 ? 'items' : 'item'}</a>`;
                table += `<ol start="0" style="display: none;">`;
                value.forEach(val => {
                    table += `<li>${val}</li>`;
                });
                table += `</ol></td>`;
            } else if (typeof value === 'object' && value !== null) {
                table += `<td>${formatAsTable(value)}</td>`;
            } else {
                table += `<td>${value !== undefined ? value : ''}</td>`;
            }
        });
        table += '</tr>';
    });
};
